import logging
import re
import numpy as np
import pandas as pd
from scipy.optimize import curve_fit
from sklearn.metrics import r2_score

def KawakitaEq(pressure, init_por, B):
    return init_por / (1 + B * pressure)

def DuckworthEq(porosity, t_hat, kb):
    return t_hat * np.exp(-kb * porosity)
        
class DataCleaning(): # for tablet masterfile

    def handle_data(self, df: pd.DataFrame) -> pd.DataFrame:
        try:
            dropped_API = df.dropna(subset=["API"])
            for i in dropped_API.index:
                if df.loc[i,"API"]:
                    match = re.match(r"([a-z]+)([0-9]+)", df.loc[i,"API"], re.I)
                    if match:
                        items = match.groups()
                        df.loc[i,"API group"] = items[0]
        
            df.loc[:,"API group"] = df.loc[:,"API group"]
            df = df.replace('nan',np.NaN)
            # sum of excipient concentrations and put it into a new column "filler conc"
            df.loc[:,["API conc", "filler 1 conc", "filler 2 conc", "excipient 3 conc", "lubricant conc", "disintegrant conc"]] = df.loc[:,["API conc", "filler 1 conc", "filler 2 conc", "excipient 3 conc", "lubricant conc", "disintegrant conc"]].fillna(value=0)
            df.loc[:,"filler conc"] = df.loc[:,"filler 1 conc"].values + df.loc[:,"filler 2 conc"].values + df.loc[:,"excipient 3 conc"].values
            # remove zero TS and POR values
            df = df.loc[~(df["Tensile strength (Mpa)"]<0.01), :]
            df = df.loc[~(df["Tablet porosity"]<0.001), :]
            
            #not nan blend code
            df = df.dropna(subset=["Blend code"])
            # sum conc ==1
            df = df.loc[df["sum conc"]==1, :]
            # no gp1 api
            df = df.loc[~(df["API"]=="gp1"), :]
                                
            df = df.drop(df[df["Source"]=="Mohammad formulation"].index)
            df = df.drop(df[df["Source"]=="UCB Data"].index)
            
            # # remove NaN compression pressures
            # compaction_data = df.loc[~df['Compression Pressure (Mpa)'].isna(), :]
            # compaction_data = compaction_data.dropna(subset=['Tablet porosity', "Tensile strength (Mpa)"])
            # # remove porosity & tensile strength outliers
            # compaction_data=compaction_data[(np.abs(stats.zscore(compaction_data['Tablet porosity'])) < 3)]
            # # recording index
            # start_recording = compaction_data.loc[compaction_data.Recording == 1,:]
            # recording_idx = start_recording.index
            
            return df         

        except Exception as e:
            logging.error(e)
            raise e
    

    def synthetise_data(self, df, r2_threshold=0.75, min_compaction_count=3, pressure_min=50, pressure_max=450, n_pressure=10):
        # Step 1: Filter blends based on the count of unique compaction_id
        blend_stats = df.groupby('Blend code').agg(
            first_index=('Compaction ID', 'idxmin'),
            last_index=('Compaction ID', 'idxmax'),
            compaction_count=('Compaction ID', 'nunique')
            ).reset_index()
        
        # Keep only blends with compaction data count >= min_compaction_count
        valid_blends = blend_stats[blend_stats['compaction_count'] >= min_compaction_count]['Blend code']
        df_filtered = df[df['Blend code'].isin(valid_blends)]
        
        # Step 2: Fit Kawakita and Duckworth equations, calculate R², and generate synthetic data
        results = []  # To store synthetic data for each blend
        
        for blend_code, group in df_filtered.groupby('Blend code'):
            pressures = group['Compression Pressure (Mpa)'].values
            porosities = group['Tablet porosity'].values
            tensile_strengths = group['Tensile strength (Mpa)'].values
            
            # Extract columns 4 to 20 (or whatever range) for the current blend
            additional_columns = group.iloc[0, 5:20] if group.shape[1] > 4 else pd.Series()  # Adjust as needed

            # Fit the Kawakita equation to the data
            try:
                popt_kawakita, _ = curve_fit(KawakitaEq, 
                                             pressures, 
                                             porosities, 
                                             bounds=([1e-10, 1e-10],[1,10000]), 
                                             p0=[0.6,100],
                                             maxfev=1000)
                init_por, B = popt_kawakita
                
                # Calculate the predicted porosity and R² for Kawakita
                porosity_pred = KawakitaEq(pressures, init_por, B)
                r2_porosity = r2_score(porosities, porosity_pred)
            except Exception:
                init_por, B, r2_porosity = np.nan, np.nan, np.nan
            
            # Fit the Duckworth equation to the porosity-tensile_strength data
            try:
                popt_duckworth, _ = curve_fit(DuckworthEq, 
                                              porosities, 
                                              tensile_strengths, 
                                              bounds=([1e-10, 1e-10],[50,50]), 
                                              p0=[10, 5],
                                              maxfev=1000)
                t_hat, kb = popt_duckworth
                
                # Calculate the predicted tensile strength and R² for Duckworth
                tensile_strength_pred = DuckworthEq(porosities, t_hat, kb)
                r2_tensile = r2_score(tensile_strengths, tensile_strength_pred)
            except Exception:
                t_hat, kb, r2_tensile = np.nan, np.nan, np.nan
            
            # Only consider blends with good quality of fit for both Kawakita and Duckworth
            if r2_porosity >= r2_threshold and r2_tensile >= r2_threshold:
                
                # Generate 7 random synthetic pressure values uniformly distributed within the given bounds
                synthetic_pressures = np.sort(np.random.uniform(pressure_min, pressure_max, size=n_pressure))
            
                # Generate synthetic porosity values using the Kawakita model
                synthetic_porosities = KawakitaEq(np.array(synthetic_pressures), init_por, B)
                
                # Generate synthetic tensile strength values using the Duckworth model
                synthetic_tensile_strengths = DuckworthEq(synthetic_porosities, t_hat, kb)
                
                # Store results (including synthetic data) in a single structure
                for i, (p, por, ts) in enumerate(zip(synthetic_pressures, synthetic_porosities, synthetic_tensile_strengths)):
                    result = {
                        'Blend code': blend_code,
                        'compaction_id': i + 1,  # Create new compaction IDs
                        'Compression Pressure (Mpa)': p,
                        'Tablet porosity': por,
                        'Tensile strength (Mpa)': ts,
                        'init_por': init_por,
                        'B': B,
                        't_hat': t_hat,
                        'kb': kb,
                        'r2_porosity': r2_porosity,
                        'r2_tensile': r2_tensile
                        }
                    
                    result.update(additional_columns.to_dict())  
                    results.append(result)
        
        # Step 3: Convert results to DataFrame
        df_synthetic = pd.DataFrame(results)
        
        return df_synthetic