import logging
import pandas as pd
import numpy as np
import os
import pickle as pk
from keras.models import load_model
from .config import Config

config = Config()

class IngestData:
    """
    Data ingestion class which ingests data from the source and returns a DataFrame.
    """

    def __init__(self) -> None:
        """Initialize the data ingestion class."""
        pass

    def get_masterfile (self, sheet_name: str, path: str, skiprows: int) -> pd.DataFrame:
        """ Reads tablet masterfile and returns as a dataframe
        
        Parameters:
            sheet_name : The sheet name of excel file containing tablet data
            path : The path to masterfile 
            skiprows: number of rows to be skipped
            
        Returns:
            df: dataframe containing tablet data
        """
        try:
            df = pd.read_excel (path,
                                sheet_name=sheet_name,
                                index_col=None,
                                header = [0],
                                skiprows = skiprows,)
            return df
        except Exception as e:
            logging.info('Failed to load data. Check if any excel file is open.')
            logging.error(e)
            raise e
    
    def scale_down(self, x, y, xnew):
        
        try:
            ynew = np.zeros_like(xnew)
            for i in range(len(xnew)):
                bin_start = i * (len(y) // len(xnew))
                bin_end = (i + 1) * (len(y) // len(xnew))
                ynew[i] = np.sum(y[bin_start:bin_end])
            return xnew, ynew
        except Exception as e:
            logging.error(e)
            raise e
    
    def get_raw_particle_data (self, descriptor: str, folder_path: str, df: pd.DataFrame) -> pd.DataFrame:
        """

        Parameters
        ----------
        descriptor : str
            DESCRPTION.
        folder_path : str
            DESCRIPTION.
        df : pd.DataFrame
            DESCRIPTION.

        Returns
        -------
        PSD_measured : TYPE
            DESCRIPTION.

        """

        indicator = "unique identifier"
        ext = (".xlsx")
        PSD_measured = pd.DataFrame()
        for ID in df.loc[:,indicator]:
            try:
                path = folder_path + str(ID) + ext
                if os.path.exists(path):
                    PSD_temp = pd.read_excel (path,
                                              sheet_name="Sheet1",
                                              index_col=None,
                                              header = [0],
                                              usecols="A:B")
                    PSD_measured[descriptor] = PSD_temp[descriptor]
                    PSD_measured[ID] = PSD_temp["%"]
                    if len(PSD_temp[descriptor]) < 1001:
                        base_x = PSD_temp[descriptor].values
                        base_ID = ID
            except:
                print(ID, " data could not be read!")
        
        if descriptor == "CE Diameter (µm)":
            PSD = pd.DataFrame()
            for ID in df.loc[:,indicator]:
                if ID in PSD_measured.columns:
                    if ID!=base_ID:
                        PSD[descriptor],  PSD[ID] = self.scale_down(PSD_measured[descriptor].values,
                                                                PSD_measured[ID].values,
                                                                base_x)
                    else:
                        PSD[ID] = PSD_measured[ID]
            del PSD_measured
            PSD_measured = PSD          
                
        return PSD_measured
    
    def get_blend_particle_data (self, descriptor: str, folder_path: str, df: pd.DataFrame, checkrawdata: bool =False):
        """
        Reads measured PSD of raw/blend data and stores them in a dataframe
        
        Inputs:
            folder_path: folder address
            df: raw material or blend master dataset (based on checkrawdata to be True or False, respecticely)
            checkrawdata: boolean to check raw (True) or blend (False) PSDs
            descriptor: specifies CE diameter (PSD) or Aspect Ratio (AR)
        Outputs:
            PSD_measured: a dataframe including the measured PSDs or ARs
            
        """
        if checkrawdata == True:
            indicator = "unique identifier"
        else:
            indicator = "Blend code"
        ext = (".xlsx")
        PSD_measured = pd.DataFrame()
        for ID in df.loc[:,indicator]:
            if checkrawdata == True:
                path = folder_path + str(ID) + ext
            else:
                path = folder_path + str(ID) + ext
            if os.path.exists(path):
                PSD_temp = pd.read_excel (path,
                                          sheet_name="Sheet1",
                                          index_col=None,
                                          header = [0],
                                          usecols="A:B")
                PSD_measured[descriptor] = PSD_temp[descriptor]
                PSD_measured = pd.concat([PSD_measured, PSD_temp["%"].rename(ID)], axis=1)
        return PSD_measured
    
    def load_model(self, 
                   folder_path: str, 
                   model_name: str, 
                   model_suffix: str):
        """
        load models and data (n_ensembles, pkl or h5)
        choose each model to load in steps
        
        class formulation & process : mixture model methods and
        process methods should be done in steps, so we can flexibly choose 
        which models should be loaded and simulated
        
        models should be added as a feature to the hybrid machine class
        each method for a model should be called if the model of that method exists
        
        add model: load it, add it to hybrid machine

        Returns
        -------
        None.

        """     

        if model_suffix == ".pkl":
            model = pk.load(open(folder_path + model_name + model_suffix,'rb'))
        elif model_suffix == ".h5":
            model = []
            for i in range(config.n_ensembles):
                single_model = load_model(os.path.join(folder_path,
                                                f"{model_name}_{i}{model_suffix}")
                                   )
                model.append(single_model)
        
        return model