import os
import pickle as pk
import warnings
import xgboost as xgb
import random
import pandas as pd
import numpy as np
from .config import Config
from .data_visualisation import (
    show_pc_scores_2d,
    get_generic_ID, 
    get_generic_blendcode, 
    )
from typing import Union
import scipy
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
from sklearn.model_selection import GroupShuffleSplit
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
import tensorflow as tf
from tensorflow import keras 
from keras import layers
from keras import callbacks
from keras.models import save_model
from keras.optimizers import Adam
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.pyplot import figure
import seaborn as sns

# fix random seed
random.seed(1)
np.random.seed(42)
    
# future warnings
warnings.filterwarnings("ignore")
warnings.simplefilter(action='ignore', category=FutureWarning)
warnings.simplefilter(action='ignore', category=DeprecationWarning)
warnings.simplefilter(action='ignore', category=UserWarning)

config = Config()

def active_components (df_row: Union[pd.DataFrame, pd.Series, list]) -> Union[pd.DataFrame, pd.Series, list] :
    # df_row: series of components separated by the component titles (type: a row of pandas)
    # removes "xxx" placebo/empty components 
    # Example: active_components (df_comp_model_params.loc[380,comp_list])
    
    # active_comp_list = df_row.values.tolist()
    if isinstance(df_row, pd.DataFrame):
        active_comp_list = df_row.values.tolist()
    elif isinstance(df_row, list):
        active_comp_list = df_row
    else:
        active_comp_list = df_row
        
    while "xxx" in active_comp_list:
        active_comp_list.remove("xxx")
    
    return active_comp_list

def replace_with_alternatives(idx: str) -> str:
    alternative_raw_dict = config.alternative_raw_dict
    if idx in alternative_raw_dict.keys():
       # print("replaced ", idx, " with ", alternative_raw_dict[idx])
       idx =  alternative_raw_dict[idx]
    return idx

def raw_components_check (df_row, raw_list) -> bool:
    # df_row: series of components separated by the component titles (type: a row of pandas)
    # raw_list: list of existing raw material components, e.g. raw TD or raw PSD (type: list)
    # returns True if all components of the blends exist in the raw component dataset
    # Example: print(raw_components_check (df_comp_model_params.loc[380,comp_list], PSD_raw.columns))
    
    # alternative_raw_dict = config.alternative_raw_dict

    list1 = active_components(df_row)
    for i, component in enumerate(list1):
        # if component in alternative_raw_dict.keys():
        #     list1[i] =  alternative_raw_dict[component]
        list1[i] = replace_with_alternatives(component)

    return set(list1).issubset(set(raw_list))

def create_dict (df_raw: pd.DataFrame, quantity:str) -> dict:
    """ 
    Creates dictionary of raw components and their quantity
    
    Parameters:
        df_raw: dataframe of raw masterfile
        quantity: true or bulk density
        
    Returns:
        dict_raw: dictionary containing raw components and their corresponding quantity values
    """
    temp = df_raw.loc[~df_raw[quantity].isna(),["unique identifier", quantity]]
    dict_raw = dict(zip(temp["unique identifier"], temp[quantity]))
    # correct TD of la6
    if quantity=="True Density":
        dict_raw["la6"] = 1.54 # 1.0929061829875715
    del temp
    return dict_raw

class TrueDensityModel:

    def TrueDensitySingle(self, components, fractions, dict_td):
        """
        Uses a single row of formulation data and calculates true density of that formulation
        
        Inputs:
            components: list of components in the formulation
            fractions: mass fractions of components
            dict_td: dictionary of raw material true density
        Outputs:
            true_density: true density of that specific row/formulation (single value)
        """
        sum_conc = 0
        TD_mix_inv = 0
        for i, component in enumerate(components):
            idx = component
            idx = replace_with_alternatives(idx)
            if idx in dict_td.keys():
                TD_idx = dict_td[idx]
                TD_mix_inv = TD_mix_inv + (fractions[i] / TD_idx)
                # TD_mix_inv = TD_mix_inv + (df_blend_TD.loc[row, [component + " conc"]].values * np.log(TD_idx))
                sum_conc += fractions[i]
            else:
                TD_mix_inv = TD_mix_inv + 0
                sum_conc +=0
    # what if an idx is not existant in the dict_td? just considers zero now!
        if TD_mix_inv!=0:
            true_density = sum_conc/TD_mix_inv
            # true_density = np.exp(float(TD_mix_inv))
        else:
            true_density = TD_mix_inv
        return float(true_density)


    def TrueDensity(self, df_blend_TD, row, blend_comp_list, dict_td):
        #TODO: merge this with the function above
        """
        Uses a single row of formulation data and calculates true density of that formulation
        
        Inputs:
            df_blend_TD: dataframe including component types and concentrations
            row: specifies the row of df_blend_TD to be used for true density mixture calculations
            blend_comp_list: list of components in the blend
            dict_td: dictionary of raw material true density
        Outputs:
            true_density: true density of that specific row/formulation (single value)
        """
        sum_conc = 0
        TD_mix_inv = 0
        for component in blend_comp_list:
            idx = df_blend_TD.loc[row, component]
            idx = replace_with_alternatives(idx)
            if idx in dict_td.keys():
                TD_idx = dict_td[idx]
                TD_mix_inv = TD_mix_inv + (df_blend_TD.loc[row, [component + " conc"]].values / TD_idx)
                # TD_mix_inv = TD_mix_inv + (df_blend_TD.loc[row, [component + " conc"]].values * np.log(TD_idx))
                sum_conc += df_blend_TD.loc[row, [component + " conc"]].values
            else:
                TD_mix_inv = TD_mix_inv + 0
                sum_conc +=0
    # what if an idx is not existant in the dict_td? just considers zero now!
        if TD_mix_inv!=0:
            true_density = sum_conc/TD_mix_inv
            # true_density = np.exp(float(TD_mix_inv))
        else:
            true_density = TD_mix_inv
        return float(true_density)
    
    
    def predict_tablet(self, df_raw, df_tablet):
        """
        Uses tablet master file and raw material true densities to calculate mixture true density of tablets
        
        Inputs:
            df: tablet master file
            dict_td: dictionary of raw material true densities
            
        outputs:
            added column "Mixture True Density" in df (talet master file)
            
        """
        dict_td = create_dict(df_raw, "True Density")
        # use the dict of true density of each component to calculate true density of a mixture of several components
        # returns df.loc[:, "Mixture True Density"]
        comp_list = ["API", "disintegrant", "filler 1", "filler 2", "excipient 3", "lubricant"]
        for row in df_tablet.index:
            TD_raw_check = raw_components_check (df_tablet.loc[row,comp_list], dict_td.keys())
            if TD_raw_check:
                df_tablet.loc[row, "Mixture True Density"] = self.TrueDensity(df_tablet, row, comp_list, dict_td)
        return df_tablet
    
    
    def predict_blend(self, df_raw, df_blend):
        """
        Uses blend master file and raw material true densities to calculate mixture true density of blends
        
        Inputs:
            df_blend: blend master file
            dict_td: dictionary of raw material true densities
            
        outputs:
            df_blend_TD: an added column "True Density Model" 
                        
        """    
        dict_td = create_dict(df_raw, "True Density")
        blend_comp_list = config.blend_comp_list

        for row in df_blend.index:
            TD_raw_check = raw_components_check (df_blend.loc[row,blend_comp_list], dict_td.keys())
            if TD_raw_check:
                df_blend.loc[row, "True Density Model"] = self.TrueDensity(df_blend, row, blend_comp_list, dict_td)
        return df_blend


class BulkDensityModel:

    
    def BulkDensitySingle(self, components, fractions, dict_td, dict_bd):

        top, bottom = 0, 0

        for i, component in enumerate(components):
            idx = component
            idx = replace_with_alternatives(idx)
            if idx != "xxx" and idx!="nan":
                TD_idx = dict_td[idx]
                BD_idx = dict_bd[idx]
                c = fractions[i]
                top = top + (c * BD_idx / TD_idx)
                bottom = bottom + c / TD_idx
        return float(top/bottom)
    
    def BulkDensity(self, df_blend, row, blend_comp_list, dict_td, dict_bd):
        #TODO: merge this with the function above
        """
        Uses a single row of formulation data and calculates bulk density of that formulation
        
        Inputs:
            df_blend_TD: dataframe including component types and concentrations
            row: specifies the row of df_blend_TD to be used for true density mixture calculations
            blend_comp_list: list of components in the blend
            dict_td: dictionary of raw material true density
        Outputs:
            true_density: true density of that specific row/formulation (single value)
        """
        top, bottom = 0, 0
    
        for component in blend_comp_list:
            idx = df_blend.loc[row, component]
            idx = replace_with_alternatives(idx)
            if idx != "xxx" and idx!="nan":
                # TD_idx = dict_td[idx]
                BD_idx = dict_bd[idx]
                c = df_blend.loc[row, [component + " conc"]].values
                top = top + (c * BD_idx) #/ TD_idx)
                bottom = bottom + c #/TD_idx
        return float(top/bottom) 
    
    
    def predict_tablet(self, df_raw, df_tablet):
        """
        Uses tablet master file and raw material true densities to calculate mixture true density of tablets
        
        Inputs:
            df: tablet master file
            dict_td: dictionary of raw material true densities
            
        outputs:
            added column "Mixture True Density" in df (talet master file)
            
        """
        dict_td = create_dict(df_raw, "True Density")
        dict_bd = create_dict(df_raw, "BD")
        # use the dict of true density of each component to calculate true density of a mixture of several components
        # returns df.loc[:, "Mixture True Density"]
        comp_list = ["API", "disintegrant", "filler 1", "filler 2", "excipient 3", "lubricant"]
        for row in df_tablet.index:
            TD_raw_check = raw_components_check (df_tablet.loc[row,comp_list], dict_td.keys())
            if TD_raw_check:
                df_tablet.loc[row, "Mixture Bulk Density"] = self.BulkDensity (df_tablet, row, comp_list, dict_td, dict_bd)
        return df_tablet
    
    
    def predict_blend(self, df_raw, df_blend):
        """
        Uses blend master file and raw material true densities to calculate mixture true density of blends
        
        Inputs:
            df_blend: blend master file
            dict_td: dictionary of raw material true densities
            
        outputs:
            df_blend_TD: an added column "True Density Model" 
                        
        """    
        dict_td = create_dict(df_raw, "True Density")
        dict_bd = create_dict(df_raw, "BD")
        blend_comp_list = config.blend_comp_list
        for row in df_blend.index:
            TD_raw_check = raw_components_check (df_blend.loc[row,blend_comp_list], dict_td.keys())
            BD_raw_check = raw_components_check (df_blend.loc[row,blend_comp_list], dict_bd.keys())
            if TD_raw_check & BD_raw_check:
                df_blend.loc[row, "BD Model"] = self.BulkDensity (df_blend, row, blend_comp_list, dict_td, dict_bd)
        return df_blend


class ParticleModel:
    
    def PSD_predictSingle (self, components, fractions, PSD_raw, dict_td):
        """
        receives one row of formulation (e.g. in tablet or blend masterfile) and calculates the PSD of the mixture
        Same approach as in "PSD_probabilistic_number_based_model" function, but uses one row (i.e. formulation) only
        
        """

        existing_raw_PSD = PSD_raw.columns.values.tolist()
        existing_raw_PSD.append("xxx")
        a, n_comp = 0, 0
        for item in components:
            item = replace_with_alternatives(item)
            if item in existing_raw_PSD:
                a = a + 1
                if item != "xxx":
                    n_comp = n_comp + 1
        if a == len(components):
            S = np.zeros(n_comp)
            # shape_factor = np.zeros(n_comp)
            b = 0
            for i, component in enumerate(components):
                idx = component
                idx = replace_with_alternatives(idx)
                if idx != "xxx":
                    idx_conc = fractions[i]
                    temp = PSD_raw.iloc[:, 0]**-3
                    S[b] = ((temp.sum())*(idx_conc))/dict_td[idx]
                    b = b + 1
            S = S/np.sum(S)
            weighted_average = pd.DataFrame()
            weighted_average["WA"] = pd.DataFrame(np.zeros((len(PSD_raw),1)))
            c = 0 
            for component in components:
                idx = component
                idx = replace_with_alternatives(idx)
                if idx != "xxx":
                    weighted_average["WA"] += (PSD_raw.loc[:, idx]*S[c])
                    c = c + 1
        return weighted_average["WA"].to_numpy()
    
    def PSD_predict (self, df_blend, row, blend_comp_list, PSD_raw,  dict_td):
        #TODO: merge this with the function above
        """
        receives one row of formulation (e.g. in tablet or blend masterfile) and calculates the PSD of the mixture
        Same approach as in "PSD_probabilistic_number_based_model" function, but uses one row (i.e. formulation) only
        
        """
        # TODO: two blend comp list for blend and tablet master files, then run the model for each row
        existing_raw_PSD = PSD_raw.columns.values.tolist()
        existing_raw_PSD.append("xxx")

        a, n_comp = 0, 0
        for item in df_blend.loc[row, blend_comp_list]:
            item = replace_with_alternatives(item)
            if item in existing_raw_PSD:
                a = a + 1
                if item != "xxx":
                    n_comp = n_comp + 1
        if a == len(blend_comp_list):
            S = np.zeros(n_comp)
            # shape_factor = np.zeros(n_comp)
            b = 0
            for component in blend_comp_list:
                idx = df_blend.loc[row, component]
                idx = replace_with_alternatives(idx)
                if idx != "xxx":
                    idx_conc = df_blend.loc[row, [component + " conc"]]
                    temp = PSD_raw.iloc[:, 0]**-3
                    S[b] = ((temp.sum())*(idx_conc))/dict_td[idx]
                    b = b + 1
            S = S/np.sum(S)
            weighted_average = pd.DataFrame()
            weighted_average["WA"] = pd.DataFrame(np.zeros((len(PSD_raw),1)))
            c = 0 
            for component in blend_comp_list:
                idx = df_blend.loc[row, component]
                idx = replace_with_alternatives(idx)
                if idx != "xxx":
                    weighted_average["WA"] += (PSD_raw.loc[:, idx]*S[c])
                    c = c + 1
    
        return weighted_average["WA"]
    
    
    def AR_predictSingle (self, components, fractions, PSD_raw, dict_td):
        """
        receives one row of formulation (e.g. in tablet or blend masterfile) and calculates the AR of the mixture
        Same approach as in "PSD_probabilistic_weighted_average" function, but uses one row (i.e. formulation) only
        
        """
        truedensitymodel = TrueDensityModel()
        # AR probabilistic mixture model using weighted average method
        existing_raw_PSD = PSD_raw.columns.values.tolist()
        existing_raw_PSD.append("xxx")
        a, n_comp = 0, 0
        for item in components:
            item = replace_with_alternatives(item)
            if item in existing_raw_PSD:
                a = a + 1
                if item != "xxx":
                    n_comp = n_comp + 1
        if a == len(components):
            weighted_average = pd.DataFrame()
            weighted_average["WA"] = pd.DataFrame(np.zeros((len(PSD_raw),1)))
            TD_mix = truedensitymodel.TrueDensitySingle(components, fractions, dict_td)
            for i, component in enumerate(components):
                idx = component
                idx = replace_with_alternatives(idx)
                if idx != "xxx" and idx!="nan":
                    idx_conc = fractions[i]
                    weighted_average["WA"] += (PSD_raw.loc[:, idx]*float(idx_conc)*(TD_mix/dict_td[idx]))
                else:
                    idx_conc = 0

        return  weighted_average["WA"].to_numpy()
    
    def AR_predict (self, df_blend, row, blend_comp_list, PSD_raw, dict_td):
        #TODO: merge this with the function above
        """
        receives one row of formulation (e.g. in tablet or blend masterfile) and calculates the AR of the mixture
        Same approach as in "PSD_probabilistic_weighted_average" function, but uses one row (i.e. formulation) only
        
        """
        # AR probabilistic mixture model using weighted average method
        existing_raw_PSD = PSD_raw.columns.values.tolist()
        existing_raw_PSD.append("xxx")
        a, n_comp = 0, 0
        for item in df_blend.loc[row, blend_comp_list]:
            item = replace_with_alternatives(item)
            if item in existing_raw_PSD:
                a = a + 1
                if item != "xxx":
                    n_comp = n_comp + 1
        if a == len(blend_comp_list):
            weighted_average = pd.DataFrame()
            weighted_average["WA"] = pd.DataFrame(np.zeros((len(PSD_raw),1)))
            truedensitymodel=TrueDensityModel()
            TD_mix = truedensitymodel.TrueDensity(df_blend, row, blend_comp_list, dict_td)
            for component in blend_comp_list:
                idx = df_blend.loc[row, component]
                idx = replace_with_alternatives(idx)
                if idx != "xxx" and idx!="nan":
                    idx_conc = df_blend.loc[row, [component + " conc"]]
                    weighted_average["WA"] += (PSD_raw.loc[:, idx]*float(idx_conc.values)*(TD_mix/dict_td[idx]))
                else:
                    idx_conc = 0

        return  weighted_average["WA"]
    
    def PSD_probabilistic_number_based_model (self, df_raw, df_blend, PSD_raw, blend_comp_list):
        """
        * works without raw components check
        Uses blend formulations in blend masterfile and raw PSDs to calculate mixture PSDs
        Methodology is based on the the size classes, raw PSDs, true density, and concentrations
        """
        dict_td = create_dict(df_raw, "True Density")
        blend_comp_list = blend_comp_list
        df_blend_model = pd.DataFrame()
        descriptor = PSD_raw.columns[0]
        df_blend_model[descriptor] = PSD_raw[descriptor]
        existing_raw_PSD = PSD_raw.columns.values.tolist()
        existing_raw_PSD.append("xxx")
        for row in df_blend.index:
            blend_code = df_blend.loc[row, "Blend code"]
            a, n_comp = 0, 0
            for item in df_blend.loc[row, blend_comp_list]:
                if item in existing_raw_PSD:
                    a = a + 1
                    if item != "xxx":
                        n_comp = n_comp + 1
            if a == len(blend_comp_list):
                S = np.zeros(n_comp)
                # shape_factor = np.zeros(n_comp)
                b = 0
                for component in blend_comp_list:
                    idx = df_blend.loc[row, component]
                    idx = replace_with_alternatives(idx)
                    if idx != "xxx":
                        idx_conc = df_blend.loc[row, [component + " conc"]]
                        # if idx == "la1":
                        #     idx_conc = idx_conc * 0.9
                        temp = PSD_raw.loc[:, descriptor]**(-3)
                        S[b] = ((temp.sum())*(idx_conc))/dict_td[idx]
                        b = b + 1
                S = S/np.sum(S)
                weighted_average = pd.DataFrame()
                weighted_average["WA"] = pd.DataFrame(np.zeros((len(PSD_raw),1)))
                c = 0 
                for component in blend_comp_list:
                    idx = df_blend.loc[row, component]
                    if idx != "xxx":
                        weighted_average["WA"] += (PSD_raw.loc[:, idx]*S[c])
                        c = c + 1
                df_blend_model.loc[:,blend_code] = weighted_average["WA"]
                del weighted_average     
                
        return df_blend_model
    
    
    def PSD_probabilistic_weighted_average (self, df_raw, df_blend, PSD_raw, blend_comp_list):
        """
        Uses blends formultions in blends dataset and raw PSD/AR to calculate mixture PSDs/ARs
        
        """
        #create dict true density
        dict_td = create_dict(df_raw, "True Density")
        
        #import true density method from TrueDensityModel class
        truedensitymodel = TrueDensityModel()
        
        # PSD probabilistic mixture model using weighted average method
        df_blend_model = pd.DataFrame()
        descriptor = PSD_raw.columns[0]
        df_blend_model[descriptor] = PSD_raw[descriptor]
        existing_raw_PSD = PSD_raw.columns.values.tolist()
        existing_raw_PSD.append("xxx")
        for row in df_blend.index:
            blend_code = df_blend.loc[row, "Blend code"]
            a, n_comp = 0, 0
            for item in df_blend.loc[row, blend_comp_list]:
                if item in existing_raw_PSD:
                    a = a + 1
                    if item != "xxx":
                        n_comp = n_comp + 1
            if a == len(blend_comp_list):
                weighted_average = pd.DataFrame()
                weighted_average["WA"] = pd.DataFrame(np.zeros((len(PSD_raw),1)))
                TD_mix = truedensitymodel.TrueDensity(df_blend, row, blend_comp_list, dict_td)
                for component in blend_comp_list:
                    idx = df_blend.loc[row, component]
                    if idx != "xxx" and idx!="nan":
                        idx_conc = df_blend.loc[row, [component + " conc"]]
                        weighted_average["WA"] += (PSD_raw.loc[:, idx]*float(idx_conc.values))*(TD_mix/dict_td[idx])
                    else:
                        idx_conc = 0
                        
                df_blend_model.loc[:,blend_code] = weighted_average["WA"]
                del weighted_average, idx

        return df_blend_model
    
    def particle_size_model(self, df_raw, df_blend, PSD_NB_raw, PSD_blend_NB_measured):
                
        blend_comp_list = config.blend_comp_list
        PSD_blend_NB_model_prob1 = self.PSD_probabilistic_number_based_model (df_raw, df_blend, PSD_NB_raw, blend_comp_list)
        
        temp = PSD_blend_NB_measured.copy()
        PSD_blend_NB_model_prob1_transposed = temp.drop(labels=["CE Diameter (µm)"], axis=1).T
        blends_index_psd = PSD_blend_NB_model_prob1_transposed.index
    
        PSD_Scaler = StandardScaler()
        PSD_blend_NB_model_prob1_transposed = PSD_Scaler.fit_transform(PSD_blend_NB_model_prob1_transposed)
        pcamodel_PSD = PCA(n_components=3)
        PSD_blend_principal_components = pcamodel_PSD.fit_transform(PSD_blend_NB_model_prob1_transposed)
    
        for i, blend in enumerate(blends_index_psd):
            df_blend.loc[df_blend["Blend code"]==blend, ["PSD PC1", "PSD PC2", "PSD PC3"]] = PSD_blend_principal_components[i,:]
    
        pk.dump(pcamodel_PSD, open(os.path.join("./saved_model/", f"{config._name_pca_psd}.pkl"),"wb"))
        pk.dump(PSD_Scaler, open(os.path.join("./saved_model/", f"{config._name_psd_scaler}.pkl"),"wb"))
        return PSD_blend_NB_model_prob1, PSD_Scaler, pcamodel_PSD
    
    
    def particle_shape_model(self, df_raw, df_blend, AR_raw, AR_blend_measured):
                
        blend_comp_list = config.blend_comp_list
        AR_blend_model = self.PSD_probabilistic_weighted_average (df_raw, df_blend, AR_raw, blend_comp_list)
        temp = AR_blend_measured.copy()
        AR_blend_model_transposed = temp.drop(labels=["Aspect Ratio"], axis=1).fillna(0).T
        blends_index_ar = AR_blend_model_transposed.index
    
        AR_Scaler = StandardScaler()
        AR_blend_model_transposed = AR_Scaler.fit_transform(AR_blend_model_transposed)
        pcamodel_AR = PCA(n_components=3)
        AR_blend_principal_components = pcamodel_AR.fit_transform(AR_blend_model_transposed)
    
        # for each row, put principal components into the blend dataset:
        for i, blend in enumerate(blends_index_ar):
            df_blend.loc[df_blend["Blend code"]==blend, ["AR PC1", "AR PC2", "AR PC3"]] = AR_blend_principal_components[i,:]
    
        pk.dump(pcamodel_AR, open(os.path.join("./saved_model/", f"{config._name_pca_ar}.pkl"),"wb"))
        pk.dump(AR_Scaler, open(os.path.join("./saved_model/", f"{config._name_ar_scaler}.pkl"),"wb"))
        return AR_blend_model, AR_Scaler, pcamodel_AR
    
class Model():
    """
    Base for all models in hybrid machine 
    """
    def __init__(self, model, name) -> None:
        self.obj = model
        self.name = name

    def __repr__(self):
        return (f"Model(name={self.name}")


class Formulation():
    def __init__ (self,) -> None:
        self.titles= []
        self.components= []
        self.fractions= []
        self.consolidation_pressure = 1.3
        
        self.models = dict()
        
        self.truedensitymodel = TrueDensityModel()
        self.bulkdensitymodel = BulkDensityModel()
        self.particlemodel = ParticleModel()
    
    def add_component(self, name, title, fraction) -> None: 
        self.components.append(name)
        self.titles.append(title)
        self.fractions.append(fraction)
        
        if np.sum(self.fractions) -1 > 1e-2:
            raise ValueError("Sum of fractions exceed 1.0")
        # else:
            # print(f"Component(name={name}, role={title}, fraction={fraction})")
        
    def get_data (self, df_raw, PSD_NB_raw, AR_raw, ):
        self.PSD_NB_raw = PSD_NB_raw
        self.AR_raw = AR_raw
        self.dict_trd = create_dict(df_raw, "True Density")
        self.dict_bd = create_dict(df_raw, "BD")
    
    def add_model(self, model: Model):
        self.models[model.name] = model.obj 
    
    def check_raw_method(self):
        TrD_check = raw_components_check (self.components, list(self.dict_trd.keys()))
        BD_check = raw_components_check (self.components, list(self.dict_trd.keys()))
        PSD_check = raw_components_check (self.components, self.PSD_NB_raw.columns.tolist())
        AR_check = raw_components_check (self.components, self.AR_raw.columns.tolist())
        self.check_raw = TrD_check and BD_check and PSD_check and AR_check
        return self.check_raw

    def true_density_method(self):
        self.true_density = self.truedensitymodel.TrueDensitySingle(self.components, self.fractions, self.dict_trd)
        return self.true_density

    def bulk_density_method(self):
        self.bulk_density = self.bulkdensitymodel.BulkDensitySingle(self.components, self.fractions, self.dict_trd, self.dict_bd)
        return self.bulk_density
    
    def particle_size_method(self):
        self.PSD_full_dist = self.particlemodel.PSD_predictSingle(self.components, self.fractions, self.PSD_NB_raw, self.dict_trd)
        return self.PSD_full_dist
    
    def particle_shape_method(self):
        self.AR_full_dist = self.particlemodel.AR_predictSingle(self.components, self.fractions, self.AR_raw, self.dict_trd)
        return self.AR_full_dist
    
    def PCA_PSD_method(self):
        PSD_full_dist_t = self.particle_size_method().reshape(1, -1)
        PSD_full_dist_t_sc = self.models["psd_scaler"].transform(PSD_full_dist_t)
        self.PCs_PSD = self.models["pca_psd"].transform(PSD_full_dist_t_sc)[0]
        return self.PCs_PSD
    
    def PCA_AR_method(self):
        AR_full_dist_t = self.particle_shape_method().reshape(1, -1)
        AR_full_dist_t_sc = self.models["ar_scaler"].transform(AR_full_dist_t)
        self.PCs_AR = self.models["pca_ar"].transform(AR_full_dist_t_sc)[0]
        return self.PCs_AR
    
    def tapped_density_method(self):
        td_input = np.hstack((self.PCA_PSD_method(),
                              self.PCA_AR_method(),
                              self.true_density_method(),
                              self.bulk_density_method()))
        td_input_sc = self.models["td_scaler"].transform(td_input.reshape(1,-1))
        self.tapped_density = self.models["td_model"].predict(td_input_sc)[0]
        return self.tapped_density
    
    def ffc_method(self):
        ffc_input = np.hstack((self.PCA_PSD_method(),
                               self.PCA_AR_method(),
                               self.true_density_method(),
                               self.bulk_density_method(),
                               self.consolidation_pressure))
        ffc_input_sc = self.models["ffc_scaler"].transform(ffc_input.reshape(1,-1))
        self.ffc= 1/self.models["ffc_model"].predict(ffc_input_sc)[0]
        return self.ffc
    
    def mixture_props_method(self):
        if self.check_raw_method():
            self.true_density_method()
            self.bulk_density_method()
            self.particle_size_method()
            self.particle_shape_method()
            self.PCA_PSD_method()
            self.PCA_AR_method()
            self.tapped_density_method()
            self.ffc_method()
        return self 


class Process(Formulation):
    def __init__ (self, Formulation):
        try:
            API_id = Formulation.titles.index("API")
            API_conc =  Formulation.fractions[API_id]
        except:
            API_conc = 0
            # print("API and/or its fraction not found. API fraction set to zero.")
        
        self.input_process = np.hstack((API_conc,
                                        Formulation.PCs_PSD, 
                                        Formulation.PCs_AR,
                                        Formulation.true_density,
                                        Formulation.bulk_density,))
        self.models = dict()
    
    def add_model(self, model: Model):
        self.models[model.name] = model.obj 
    
    def compaction_method(self, compression_pressure):
        self.input_process = np.hstack((self.input_process, compression_pressure)).reshape(1,-1)
        
        def porosity():
            try:
                porosity_model_input_sc = self.models["porosity_scaler"].transform(self.input_process)
            except:
                porosity_model_input_sc = self.input_process
                # print("No porosity scaler found. Running porosity with original input data.")
                
            try:
                porosity_pred = [model.predict(porosity_model_input_sc, verbose=0) for model in self.models["porosity_model"]]
            except:
                porosity_pred = [self.models["porosity_model"].predict(porosity_model_input_sc)]

            porosity_pred = np.asarray(porosity_pred)
            return porosity_pred.mean(), porosity_pred.std()
        
        self.porosity_mean, self.porosity_std = porosity()

        def tensile_strength():
            input_process_tensile = np.hstack( (self.input_process, self.porosity_mean.reshape(1,-1)))
            
            try:
                tensile_model_input_sc = self.models["tensile_scaler"].transform(input_process_tensile)
            except:
                tensile_model_input_sc = input_process_tensile
                # print("No TS scaler found. Running TS model with original input data.")
            
            try:
                tensile_pred = [model.predict(tensile_model_input_sc, verbose=0) for model in self.models["tensile_model"]]
            except:
                tensile_pred = [self.models["tensile_model"].predict(tensile_model_input_sc)]
            tensile_pred = np.asarray(tensile_pred)
            tensile_pred = np.exp(tensile_pred)
            return tensile_pred.mean(), tensile_pred.std()
            
        self.tensile_mean, self.tensile_std = tensile_strength()
        
        return self.porosity_mean, self.porosity_std, self.tensile_mean, self.tensile_std

class EmpiricalModels(): 

    def KawakitaEq(self, pressure, init_por, B):
        return init_por / ( 1 + B*pressure)

    def GurnhamEq(self, pressure, p_zero, k):
        return (-1 / k) * np.log(pressure / p_zero)

    def DuckworthEq (self, porosity, t_hat, kb):
        return t_hat * np.exp(-kb * porosity)

    def KawakitaFit(self, por, cp):
        init_params = [0.6,100]
        bounds = ([1e-10, 1e-10],[1,10000])
        opt_params, cv = scipy.optimize.curve_fit(self.KawakitaEq, cp, por, init_params, bounds = bounds)
        init_por, B = opt_params
        sigma_params = np.sqrt(np.diagonal(cv))
        return init_por, sigma_params[0], B, sigma_params[1]

    def GurnhamFit(self, por, cp):
        init_params = [500, 8]
        bounds=([1e-10, 1e-10],[2000,40])
        opt_params, cv = scipy.optimize.curve_fit(self.GurnhamEq, cp, por, init_params, bounds = bounds, maxfev = 5000)
        p_zero, k = opt_params
        sigma_params = np.sqrt(np.diagonal(cv))
        return p_zero, sigma_params[0], k, sigma_params[1]

    def DuckworthFit(self, ts, por):
        init_params = [10, 5] 
        bounds=([1e-10, 1e-10],[50,50])
        opt_params, cv = scipy.optimize.curve_fit(self.DuckworthEq, por, ts, init_params, bounds = bounds, maxfev = 5000)
        t_hat, kb = opt_params
        sigma_params = np.sqrt(np.diagonal(cv))
        return t_hat, sigma_params[0], kb, sigma_params[1]