from .ingest_data import load_models
from src.model_dev import (Model, 
                           Formulation, 
                           Process, 
                           EmpiricalModels)
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Tuple
import logging
from src.config import Config

config = Config()
def single_run(titles: list,
               components: list,
               fractions: list, 
               cp: float,):
    """
    Generates formulation and process objects for a single formulation & process

    Parameters
    ----------
    titles : list
    components : list
    fractions : list
    cp : float

    Raises
    ------
    e

    Returns
    -------
    f : TYPE
    p : TYPE

    """
    
    try:
        
        folder_path = config.saved_models_path

        psd_scaler = load_models(folder_path, "psd_scaler", ".pkl")
        ar_scaler = load_models(folder_path, "ar_scaler", ".pkl")
        pca_psd = load_models(folder_path, "pca_psd", ".pkl")
        pca_ar = load_models(folder_path, "pca_ar", ".pkl")
        td_scaler = load_models(folder_path, "td_scaler", ".pkl")
        td_model = load_models(folder_path, "td_model", ".pkl")
        ffc_scaler = load_models(folder_path, "ffc_scaler", ".pkl")
        ffc_model = load_models(folder_path, "ffc_model", ".pkl")
        # porosity_scaler = load_models(folder_path, "porosity_scaler", ".pkl")
        porosity_model = load_models(folder_path, "porosity_model", ".pkl")
        # porosity_scaler_wcsd = load_models(folder_path, "porosity_scaler_wcsd", ".pkl")
        # porosity_model_wcsd = load_models(folder_path, "porosity_model_wcsd", ".h5")
        # tensile_scaler = load_models(folder_path, "tensile_scaler", ".pkl")
        tensile_model = load_models(folder_path, "tensile_model", ".pkl")
        # tensile_scaler_wcsd = load_models(folder_path, "tensile_scaler_wcsd", ".pkl")
        # tensile_model_wcsd = load_models(folder_path, "tensile_model_wcsd", ".h5")

        df_raw = load_models(folder_path, "df_raw", ".pkl")
        psd_raw = load_models(folder_path, "psd_raw", ".pkl")
        # print(psd_raw.columns)

        ar_raw = load_models(folder_path, "ar_raw", ".pkl")
        
        f=Formulation()
        
        for name, title, fraction in zip(components, titles, fractions):
            f.add_component(name, title, fraction)
        
        f.get_data(df_raw, psd_raw, ar_raw)
        f.add_model(Model(psd_scaler, "psd_scaler"))
        f.add_model(Model(ar_scaler, "ar_scaler"))
        f.add_model(Model(pca_psd, "pca_psd"))
        f.add_model(Model(pca_ar, "pca_ar"))
        f.add_model(Model(td_scaler, "td_scaler"))
        f.add_model(Model(td_model, "td_model"))
        f.add_model(Model(ffc_scaler, "ffc_scaler"))
        f.add_model(Model(ffc_model, "ffc_model"))
        
        f.mixture_props_method()
        
        p=Process(f)
        
        # p.add_model(Model(porosity_scaler, "porosity_scaler"))
        p.add_model(Model(porosity_model, "porosity_model"))
        # p.add_model(Model(tensile_scaler, "tensile_scaler"))
        p.add_model(Model(tensile_model, "tensile_model"))
        
        p.compaction_method(compression_pressure=cp)
        
        return f, p
    
    except Exception as e:
        logging.error(e)
        raise e

def single_run_gen_profile(titles: list, 
                                components: list, 
                                fractions: list, 
                                cp_range: Tuple,
                                n_runs: int):
    """
    generates compaction profile values of a single case across different compression pressures

    Parameters
    ----------
    titles : list
    components : list
    fractions : list
    cp_range : Tuple
    n_runs : int

    Raises
    ------
    e

    Returns
    -------
    None.

    """
    try:
        cp_lb = cp_range[0]
        cp_ub = cp_range[1]
        results = pd.DataFrame()
        for i, cp in enumerate(np.linspace(cp_lb, cp_ub, n_runs)):
            f, p = single_run(titles, components, fractions, cp)
            results.loc[i, "porosity_mean"] = p.porosity_mean
            results.loc[i, "porosity_std"] = p.porosity_std
            results.loc[i, "tensile_mean"] = p.tensile_mean
            results.loc[i, "tensile_std"] = p.tensile_std
            results.loc[i, "cp"] = cp
            # print(f"{i} > Simulation run at {cp} MPa")
            
        return results
        
    except Exception as e:
        logging.error(e)
        raise e

def single_run_emp_parameters(titles: list, 
                              components: list, 
                              fractions: list, 
                              cp_range: Tuple,
                              n_runs: int):
    
    try:
        profile = single_run_gen_profile(titles, 
                                              components, 
                                              fractions, 
                                              cp_range = (90, 400),
                                              n_runs = 7)
        empiricalmodels = EmpiricalModels()
        init_por, _, B, _ = empiricalmodels.KawakitaFit(profile["porosity_mean"], profile["cp"])
        t_hat, _, kb, _ = empiricalmodels.DuckworthFit(profile["tensile_mean"], profile["porosity_mean"])    
        
        res = 0 
        
        print("------------")
        print(empiricalmodels.KawakitaEq(211, init_por, B))
        print(empiricalmodels.DuckworthEq(empiricalmodels.KawakitaEq(211, init_por, B), t_hat, kb))
        print("------------")
        
        figure, ax = plt.subplots()
        figure.dpi=300
        ax.plot(np.linspace(70, 550, 100), 1*empiricalmodels.KawakitaEq(np.linspace(70, 550, 100), init_por, B), 
                color='red', linestyle="-", label="Model Prediction (Mean)")
        ax.fill_between(np.linspace(70, 550, 100), 
                        0.83*empiricalmodels.KawakitaEq(np.linspace(70, 550, 100), init_por, B), 
                        1.17*empiricalmodels.KawakitaEq(np.linspace(70, 550, 100), init_por, B),
                        color='grey', alpha=0.3, label="95% Confidence Interval")
        plt.xlabel("Compression pressure (MPa)", fontsize=16)
        plt.ylabel("Porosity (-)", fontsize=16)
        plt.legend()
        plt.show()
        
        
        figure, ax = plt.subplots()
        figure.dpi=300
        ax.plot(np.linspace(0.045, 0.26, 100), 1*empiricalmodels.DuckworthEq(np.linspace(0.045, 0.26, 100), t_hat, kb), 
                color='red', linestyle="-", label="Model Prediction (Mean)")
        ax.fill_between(np.linspace(0.045, 0.26, 100), 
                        0.83*empiricalmodels.DuckworthEq(np.linspace(0.045, 0.26, 100), t_hat, kb), 
                        1.17*empiricalmodels.DuckworthEq(np.linspace(0.045, 0.26, 100), t_hat, kb),
                        color='grey', alpha=0.3, label="95% Confidence Interval")
        plt.xlabel("Porosity (-)", fontsize=16)
        plt.ylabel("Tensile Strength (MPa)", fontsize=16)
        plt.legend()
        plt.show()
        
        res = pd.DataFrame()
        res.loc[:, "Compression Pressure (MPa)"] = np.linspace(50, 500, 10)
        res.loc[:, "Porosity Mean (-)"] = empiricalmodels.KawakitaEq(res.loc[:, "Compression Pressure (MPa)"].values, init_por, B)
        res.loc[:, "Porosity LB (-)"] = res.loc[:, "Porosity Mean (-)"] * 0.83
        res.loc[:, "Porosity UB (-)"] = res.loc[:, "Porosity Mean (-)"] * 1.17
        res.loc[:, "Tensile Strength Mean (MPa)"] = empiricalmodels.DuckworthEq(res.loc[:, "Porosity Mean (-)"].values, t_hat, kb)
        res.loc[:, "Tensile Strength LB (MPa)"] = res.loc[:, "Tensile Strength Mean (MPa)"] * 0.83
        res.loc[:, "Tensile Strength UB (MPa)"] = res.loc[:, "Tensile Strength Mean (MPa)"] * 1.17
        res.loc[:, "Hardness Mean (N)"] = res.loc[:, "Tensile Strength Mean (MPa)"].values * 4.5 * 3.1415
        
        print(empiricalmodels.KawakitaEq(229.9, init_por, B))
        print(empiricalmodels.DuckworthEq(0.15, t_hat, kb))
        
        return (init_por, B), (t_hat, kb), res
    
    except Exception as e:
        logging.error(e)
        raise e
        
def single_run_validate(titles: list, 
                        components: list, 
                        fractions: list, 
                        cp_range: Tuple,
                        n_runs: int, 
                        valid_path: str):
    
    
    (init_por, B), (t_hat, kb) = single_run_emp_parameters(titles, components, fractions, cp_range, n_runs)
    
    file_path = valid_path
    hist_all = pd.read_pickle(file_path)
    
    empiricalmodels = EmpiricalModels()
    
    figure, ax = plt.subplots()
    figure.dpi=300
    #hist_all data points
    ax.scatter(hist_all.loc[:, 'main_comp_pressure'], hist_all.loc[:, 'porosity'], 
                c='black', label='PIBO Collected data', 
                marker='o', s=100, alpha=0.8)
    #hist_all validation point
    x = hist_all.loc[len(hist_all)-1, 'main_comp_pressure']
    y = hist_all.loc[len(hist_all)-1, 'porosity']
    ax.scatter(x, y, 
               c='blue', label='PIBO-Informed Validation point', 
               marker='o', s=100, alpha=0.8)
    # ax.plot([x, x], [0, y], color='gray', linestyle='--')  # Vertical line from (x, 0) to (x, y)
    # ax.plot([0, x], [y, y], color='gray', linestyle='--')  # Horizontal line from (0, y) to (x, y)
    #hist_all calibrated profile
    ax.plot(np.linspace(70, 550, 100), 
            empiricalmodels.KawakitaEq(np.linspace(70, 550, 100),  hist_all['K1'].iloc[-1],  hist_all['K2'].iloc[-1]), 
            color='green', linestyle="--", label="PIBO-Calibrated Kawakita Model")
    #model-predicted profile
    ax.plot(np.linspace(70, 550, 100), 1*empiricalmodels.KawakitaEq(np.linspace(70, 550, 100), init_por, B), 
            color='red', linestyle="--", label="Initial prediction of Kawakita Model")
    plt.xlabel("Compression pressure (MPa)", fontsize=12)
    plt.ylabel("Porosity (-)", fontsize=12)
    plt.legend()
    plt.show()
    
    figure, ax = plt.subplots()
    figure.dpi=300
    #hist_all data points
    ax.scatter(hist_all.loc[:, 'porosity'], hist_all.loc[:, 'tensile_strength'], 
                c='black', label='PIBO Collected data', 
                marker='o', s=100, alpha=0.8)
    #hist_all validation point
    x = hist_all.loc[len(hist_all)-1, 'porosity']
    y =  hist_all.loc[len(hist_all)-1, 'tensile_strength']
    ax.scatter(x, y, 
               c='blue', label='PIBO-Informed Validation point', 
               marker='o', s=100, alpha=0.8)
    # ax.plot([x, x], [plt.ylim()[0], y], color='gray', linestyle='--')  # Vertical line from (x, 0) to (x, y)
    # ax.plot([plt.xlim()[0], x], [y, y], color='gray', linestyle='--')  # Horizontal line from (0, y) to (x, y)
    #hist_all calibrated profile
    ax.plot(np.linspace(0.045, 0.26, 100), 
            empiricalmodels.DuckworthEq(np.linspace(0.045, 0.26, 100),  hist_all['D1'].iloc[-1],  hist_all['D2'].iloc[-1]), 
            color='green', linestyle="--", label="PIBO-Calibrated Duckworth Model")
    #model-predicted profile
    ax.plot(np.linspace(0.045, 0.26, 100), 1*empiricalmodels.DuckworthEq(np.linspace(0.045, 0.26, 100), t_hat, kb), 
            color='red', linestyle="--", label="Initial prediction of Duckworth Model")
    plt.xlabel("Porosity (-)", fontsize=12)
    plt.ylabel("Tensile Strength (MPa)", fontsize=12)
    plt.legend()
    plt.show()
        
    return hist_all


def single_run_synthetised(titles: list, 
                           components: list, 
                           fractions: list, 
                           cp: float,
                           cp_range: Tuple = (90, 400),
                           n_runs: int = 6):
    
    f, p = single_run(titles, components, fractions, cp=cp)
    
    (init_por, B), (t_hat, kb), res = single_run_emp_parameters(titles, components, fractions, cp_range, n_runs)
    
    empiricalmodels = EmpiricalModels()
    porosity = empiricalmodels.KawakitaEq(cp, init_por,  B)
    tensile_strength = empiricalmodels.DuckworthEq(cp, t_hat,  kb)
        
    return f, p, porosity, tensile_strength
        
    
    