from src.model_dev import (TrueDensityModel, 
                           BulkDensityModel, 
                           ParticleModel, 
                           TappedDensityModel, 
                           FlowabilityModel, 
                           ProcessModel)
import pandas as pd
import logging
from typing import Tuple
import sklearn


def true_density_model (df_raw: pd.DataFrame, df_blend: pd.DataFrame) -> pd.DataFrame:
    """
    runs true density model to modify blend masterfile with a new column
    
    Parameters
    ----------
    df_raw : pd.DataFrame
        DESCRIPTION.
    df_blend : pd.DataFrame
        DESCRIPTION.

    Returns
    -------
    None.

    """
    try:
        truedensitymodel = TrueDensityModel()
        df_blend = truedensitymodel.predict_blend(df_raw, df_blend)
        return df_blend
    except Exception as e:
        logging.error(e)
        raise e
        
        
def bulk_density_model (df_raw: pd.DataFrame, df_blend: pd.DataFrame) -> pd.DataFrame:
    """
    runs bulk density model to modify blend masterfile with a new column
    
    Parameters
    ----------
    df_raw : pd.DataFrame
    df_blend : pd.DataFrame

    Returns
    -------
    df_blend : pd.DataFrame

    """
    try:
        bulkdensitymodel = BulkDensityModel()
        df_blend = bulkdensitymodel.predict_blend(df_raw, df_blend)
        return df_blend
    except Exception as e:
        logging.error(e)
        raise e
        

def particle_size_model (df_raw: pd.DataFrame, 
                         df_blend: pd.DataFrame, 
                         PSD_NB_raw: pd.DataFrame, 
                         PSD_blend_NB_measured: pd.DataFrame) -> Tuple[pd.DataFrame,
                                                                      sklearn.preprocessing.StandardScaler,
                                                                      sklearn.decomposition.PCA]:
    """
    runs particle size model to create psd scaler and pca model for particle size

    Parameters
    ----------
    df_raw : pd.DataFrame
    df_blend : pd.DataFrame
    PSD_NB_raw : pd.DataFram
    PSD_blend_NB_measured : pd.DataFram

    Raises
    ------
    e

    Returns
    -------
    PSD_blend_NB_model_prob1 : TYPE
    PSD_Scaler : TYPE
    pcamodel_PSD : TYPE

    """
    
    
    try:
        particlesize = ParticleModel()
        PSD_blend_NB_model_prob1, PSD_Scaler, pcamodel_PSD = particlesize.particle_size_model(df_raw, df_blend, PSD_NB_raw, PSD_blend_NB_measured)
        return PSD_blend_NB_model_prob1, PSD_Scaler, pcamodel_PSD
    except Exception as e:
        logging.error(e)
        raise e
        

def particle_shape_model (df_raw: pd.DataFrame, 
                          df_blend: pd.DataFrame, 
                          AR_raw: pd.DataFrame, 
                          AR_blend_measured: pd.DataFrame) -> Tuple[pd.DataFrame, 
                                                                       sklearn.preprocessing.StandardScaler,
                                                                       sklearn.decomposition.PCA]:
    """
    runs particle shape model to create ar scaler and pca model for aspect ratio

    Parameters
    ----------
    df_raw : pd.DataFrame
    df_blend : pd.DataFrame
    PSD_NB_raw : pd.DataFram
    PSD_blend_NB_measured : pd.DataFram

    Raises
    ------
    e

    Returns
    -------
    AR_blend_model : TYPE
    AR_Scaler : TYPE
    pcamodel_AR : TYPE

    """
    try:
        particleshape = ParticleModel()
        AR_blend_model, AR_Scaler, pcamodel_AR = particleshape.particle_shape_model(df_raw, df_blend, AR_raw, AR_blend_measured)
        return AR_blend_model, AR_Scaler, pcamodel_AR
    except Exception as e:
        logging.error(e)
        raise e      


def tapped_density_model(df_blend: pd.DataFrame):
    """
    trains tapped density model and returns the standard scaler and the model

    Parameters
    ----------
    df_blend : TYPE

    Raises
    ------
    e

    Returns
    -------
    td_scaler : TYPE
    td_model : TYPE

    """
    
    try:
        tdmodel = TappedDensityModel()
        td_scaler, td_model = tdmodel.train_model(df_blend)
        return td_scaler, td_model
    except Exception as e:
        logging.error(e)
        raise e     
        

def ffc_model(df_blend: pd.DataFrame, df_raw: pd.DataFrame = None):
    """
    trains ffc model and returns the standard scaler and the model

    Parameters
    ----------
    df_blend : pd.DataFrame

    Raises
    ------
    e

    Returns
    -------
    ffc_scaler : TYPE
    ffc_model : TYPE

    """
    
    try:
        fmodel = FlowabilityModel()
        ffc_scaler, model_ffc = fmodel.train_ffc_model(df_blend, df_raw)
        return ffc_scaler, model_ffc
    except Exception as e:
        logging.error(e)
        raise e   


def eaoif_model(df_blend: pd.DataFrame, df_raw: pd.DataFrame = None):
    """
    trains ffc model and returns the standard scaler and the model

    Parameters
    ----------
    df_blend : pd.DataFrame

    Raises
    ------
    e

    Returns
    -------
    ffc_scaler : TYPE
    ffc_model : TYPE

    """
    
    try:
        fmodel = FlowabilityModel()
        ffc_scaler, model_ffc = fmodel.train_eaoif_model(df_blend, df_raw)
        return ffc_scaler, model_ffc
    except Exception as e:
        logging.error(e)
        raise e   

def porosity_model(df_tablet: pd.DataFrame, 
                   df_raw: pd.DataFrame, 
                   PSD_NB_raw: pd.DataFrame, 
                   AR_raw: pd.DataFrame,
                   get_csd_features:bool = False):
    """
    trains porosity model

    Parameters
    ----------
    df_tablet : pd.DataFrame
    df_raw : pd.DataFrame
    PSD_NB_raw : pd.DataFrame
    AR_raw : pd.DataFrame

    Raises
    ------
    e

    Returns
    -------
    message : Boolean

    """
    
    try:
        model = ProcessModel()
        df_tablet_summary = model.porosity_model_train(df_tablet, 
                                                       df_raw, 
                                                       PSD_NB_raw, 
                                                       AR_raw, 
                                                       get_csd_features)
        return df_tablet_summary
    except Exception as e:
        logging.error(e)
        raise e   
        
        
def tensile_model(df_tablet: pd.DataFrame, 
                  df_raw: pd.DataFrame, 
                  PSD_NB_raw: pd.DataFrame, 
                  AR_raw: pd.DataFrame,
                  get_csd_features:bool = False):
    """
    trains tensile strength model

    Parameters
    ----------
    df_tablet : pd.DataFrame
    df_raw : pd.DataFrame
    PSD_NB_raw : pd.DataFrame
    AR_raw : pd.DataFrame

    Raises
    ------
    e

    Returns
    -------
    message : Boolean

    """
    
    try:
        model = ProcessModel()
        df_tablet_summary = model.tensile_model_train(df_tablet,
                                                      df_raw, 
                                                      PSD_NB_raw, 
                                                      AR_raw, 
                                                      get_csd_features)
        return df_tablet_summary
    except Exception as e:
        logging.error(e)
        raise e   